@@ -21,9 +21,9 @@
 # See the file "license.terms" for information on usage and redistribution
 # of this file, and for a DISCLAIMER OF ALL WARRANTIES.
 #
-# RCS: @(#) $Id: httpd.tcl,v 1.51 2000/08/02 07:25:53 welch Exp $
+# RCS: @(#) $Id: httpd.tcl,v 1.68 2001/07/12 01:12:21 welch Exp $
 
-package provide httpd 1.3
+package provide httpd 1.5
 
 # initialize all the global data
 
@@ -43,6 +43,7 @@ array set Httpd_Errors {
     411 {Length Required}
     419 {Expectation Failed}
     500 {Server Internal Error}
+    501 {Server Busy}
     503 {Service Unavailable}
     504 {Service Temporarily Unavailable}
 }
@@ -68,6 +69,52 @@ array set Httpd_EnvMap {
     AUTH_TYPE		auth_type
 }
 
+# The per-connection state is kept in the "data" array, which is
+# really an array named Httpd$sock - (note the upvar #0 trick throughout)
+# The elements of this array are documented here.  URL implementations
+# are free to hang additional state off the data array so long as they
+# do not clobber the elements documented here:
+
+# These fields are semi-public, or "well known".  There are a few
+# API's to access them, but URL implementations can rely on these:
+#
+# self		A list of protocol (http or https), name, and port that
+#		capture the server-side of the socket address
+#		Available with  Httpd_Protocol, Httpd_Name, and Httpd_Port API.
+# uri		The complete URL, including proto, servername, and query
+# proto		http or https
+# url		The URL after the server name and before the ?
+# query		The URL after the ?
+# ipaddr	The remote client's IP address
+# cert		Client certificate (The result of tls::status)
+# host		The host specified in the URL, if any (proxy case)
+# port		The port specified in the URL, if any
+# mime,*	HTTP header request lines (e.g., mime,content-type)
+# count		Content-Length
+# set-cookie	List of Set-Cookie headers to stick into the response
+#		Use Httpd_SetCookie to append to this.
+
+# prefix	(Set by Url_Dispatch to be the URL domain prefix)
+# suffix	(Set by Url_Dispatch to be the URL domain suffix)
+
+# auth_type	(Set by the auth.tcl module to "Basic", etc.)
+# remote_user	(Set by the auth.tcl to username from Basic authentication)
+# session	(Set by the auth.tcl to "realm,$username" from Basic auth)
+#		You can overwrite this session ID with something more useful
+
+# Internal fields used by this module.
+# left		The number of keep-alive connections allowed
+# cancel	AfterID of event that will terminate the connection on timeout
+# state		State of request processing
+# version	1.0 or 1.1
+# line		The current line of the HTTP request
+# mimeorder	List indicating order of MIME header lines
+# key		Current header key
+# checkNewLine	State bit for Netscape SSL newline bug hack
+# callback	Command to invoke when request has completed
+# file_size	Size of file returned by ReturnFile
+# infile	Open file used by fcopy to return a file, or CGI pipe
+
 # Httpd_Init
 #	Initialize the httpd module.  Call this early, before Httpd_Server.
 #
@@ -109,33 +156,6 @@ proc Httpd_Init {} {
     }
     Httpd_Version
     append Httpd(server) $Httpd(version)
-
-    # TODO - move these OUT of this file and into the main startup code
-
-    Counter_Init
-    Mtype_ReadTypes [file join $Httpd(library) mime.types]
-
-    # SSL Support - These are just the run-time defaults
-
-    set Httpd(SSL_REQUEST) 1    ;# dot request a cert
-    set Httpd(SSL_REQUIRE) 0    ;# don't require a cert
-
-    # File containing Server Certificate
-
-    set Httpd(SSL_CERTFILE) $Httpd(library)/server.pem
-    set Httpd(SSL_KEYFILE) ""
-
-    # Dir/File containing CA's
-
-    set Httpd(SSL_CADIR) $Httpd(library)
-    set Httpd(SSL_CAFILE) server.pem
-
-    # For SSL2/3 + RSA you need RSA-enabled versions of OpenSSL
-
-    set Httpd(USE_SSL2) 1
-    set Httpd(USE_SSL3) 1
-    set Httpd(USE_TLS1) 0
-    set Httpd(SSL_CIPHERS) ""        ;# use defaults
 }
 
 # Httpd_Server --
@@ -178,7 +198,6 @@ proc Httpd_Server {{port 80} {name {}} {ipaddr {}}} {
     if {[catch $cmd Httpd(listen)]} {
         return -code error "$Httpd(name):$port $Httpd(listen)\ncmd=$cmd"
     }
-    Counter_Reset accepts
 }
 
 # Httpd_SecureServer --
@@ -212,30 +231,14 @@ proc Httpd_SecureServer {{port 443} {name {}} {ipaddr {}}} {
     if {[string length $name] == 0} {
 	set Httpd(name) [info hostname]
     }
-    if {[info exists Httpd(port)] == 0} {
-	set Httpd(port) $port
-    }
     package require tls
 
-    if {![file exists $Httpd(SSL_CADIR)] && ![file exists $Httpd(SSL_CAFILE)]} {
-	return -code error "Need a CA directory or a CA file: \
-		file \"$Httpd(SSL_CAFILE)\" not found"
-    }
-    if {![file exists $Httpd(SSL_CERTFILE)]} {
-	return -code error "Certificate  \"$Httpd(SSL_CERTFILE)\" not found"
-    }
+    # This now depends on a call to tls::init being made elsewhere, typically
+    # in the main startup script.  That call sets all the various SSL parameters
+    # based on the server's configuration file.
+
     set cmd [list tls::socket -server [list HttpdAccept \
 	    [list https $name $port]]]
-    lappend cmd -request $Httpd(SSL_REQUEST) \
-	    -require $Httpd(SSL_REQUIRE) \
-	    -ssl2 $Httpd(USE_SSL2) \
-	    -ssl3 $Httpd(USE_SSL3) \
-	    -tls1 $Httpd(USE_TLS1) \
-	    -cipher $Httpd(SSL_CIPHERS) \
-	    -cadir $Httpd(SSL_CADIR) \
-	    -cafile $Httpd(SSL_CAFILE) \
-	    -certfile $Httpd(SSL_CERTFILE) \
-	    -keyfile $Httpd(SSL_KEYFILE)
     if {[string length $ipaddr] != 0} {
         lappend cmd -myaddr $ipaddr
     }
@@ -243,7 +246,6 @@ proc Httpd_SecureServer {{port 443} {name {}} {ipaddr {}}} {
     if {[catch $cmd Httpd(https_listen)]} {
         return -code error "$Httpd(name):$port $Httpd(https_listen)\ncmd=$cmd"
     }
-    Counter_Reset accepts
 }
 
 # Httpd_Shutdown --
@@ -328,7 +330,7 @@ proc HttpdAccept {self sock ipaddr port} {
 	# We do that by calling tls::handshake in a fileevent
 	# until it is complete, or an error occurs.
 
-	set data(ssl) 1
+	Count accept_https
 	fconfigure $sock -blocking 0
 	fileevent $sock readable [list HttpdHandshake $sock]
     } else {
@@ -426,26 +428,49 @@ proc HttpdReset {sock {left {}}} {
 
     # Set up a timer to close the socket if the next request
     # is not completed soon enough.  The request has already
-    # been started, but a bad client might not finish.
+    # been started, but a bad URL domain might not finish.
 
     set data(cancel) [after $Httpd(timeout1) \
-	[list Httpd_SockClose $sock 1 ""]]
+	[list Httpd_SockClose $sock 1 "timeout"]]
     fconfigure $sock -blocking 0 -buffersize $Httpd(bufsize) \
 	-translation {auto crlf}
     fileevent $sock readable [list HttpdRead $sock]
     fileevent $sock writable {}
 }
 
+# Httpd_Peername --
+#
 # Really need to fix the core to support DNS lookups.
 # This routine is not used anywhere.
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	The clients dns name.
+#
+# Side Effects:
+#	None
 
 proc Httpd_Peername {sock} {
     # This is expensive!
     fconfigure $sock -peername
 }
 
+# HttpdRead --
+#
 # Read request from a client.  This is the main state machine
 # for the protocol.
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Reads the request from the socket and dispatches the
+#	URL request when ready.
 
 proc HttpdRead {sock} {
     global Httpd
@@ -544,7 +569,7 @@ proc HttpdRead {sock} {
 	    if {$data(proto) == "POST"} {
 		fconfigure $sock  -translation {binary crlf}
 		if {![info exists data(mime,content-length)]} {
-		    Httpd_Error $sock 411 $data(mime,expect)
+		    Httpd_Error $sock 411
 		    return
 		}
 		set data(count) $data(mime,content-length)
@@ -591,6 +616,16 @@ proc HttpdRead {sock} {
 		    Httpd_Error $sock 400 "No proxy support\n$err"
 		}
 	    } else {
+		# Dispatch to the URL implementation.
+
+		# As a service for domains that loose track of their
+		# context (e.g., .tml pages) we save the socket in a global.
+		# If a domain implementation would block and re-enter the
+		# event loop, it must use Httpd_Suspend to clear this state,
+		# and use Httpd_Resume later to restore it.
+
+		set Httpd(currentSocket) $sock
+		CountStart serviceTime $sock
 		Url_Dispatch $sock
 	    }
 	}
@@ -646,12 +681,77 @@ proc Httpd_GetPostData {sock varName {size -1}} {
     return $data(count)
 }
 
+# Httpd_ReadPostDataAsync --
+#
+#	Convenience layer on Http-GetPostDataAsync to
+#	read the POST data into a the data(query) variable.
+#
+# Arguments:
+#	(Same as HttpdReadPost)
+#
+# Side Effects:
+#	(See Httpd_GetPostDataAsync)
+
+proc Httpd_ReadPostDataAsync {sock cmd} {
+    global Httpd
+    upvar #0 Httpd$sock data
+    if {[string length $data(query)]} {
+	# This merges query data from the GET/POST URL
+	append data(query) &
+    }
+    Httpd_Suspend $sock
+    fileevent $sock readable [list HttpdReadPostGlobal $sock \
+	    Httpd${sock}(query) $Httpd(bufsize) $cmd]
+    return
+}
+
 # Httpd_GetPostDataAsync --
 #
 #	Read the POST data into a Tcl variable, but do it in the
 #	background so the server doesn't block on the socket.
 #
 # Arguments:
+#	(Same as HttpdReadPost)
+#
+# Side Effects:
+#	This schedules a readable fileevent to read all the POST data
+#	asynchronously.  The data is appened to the named variable.
+#	The callback is made 
+
+proc Httpd_GetPostDataAsync {sock varName blockSize cmd} {
+    Httpd_Suspend $sock
+    fileevent $sock readable \
+	[list HttpdReadPostGlobal $sock $varName $blockSize $cmd]
+    return
+}
+
+# HttpdReadPostGlobal --
+#
+# This fileevent callback can only access a global variable.
+# But HttpdReadPost needs to affect a local variable in its
+# caller so it can be shared with Httpd_GetPostData.
+# So, the fileevent case has an extra procedure frame.
+#
+# Arguments:
+#	(Same as HttpdReadPost)
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Accumulates POST data into the named variable
+
+proc HttpdReadPostGlobal {sock varName blockSize {cmd {}}} {
+    upvar #0 $varName buffer
+    HttpdReadPost $sock buffer $blockSize $cmd
+}
+
+# HttpdReadPost --
+#
+#	The core procedure that reads post data and accumulates it
+#	into a Tcl variable.
+#
+# Arguments:
 #	sock	Client connection
 #	varName	Name of buffer variable to append post data to.  This
 #		must be a global or fully scoped namespace variable, or
@@ -663,27 +763,11 @@ proc Httpd_GetPostData {sock varName {size -1}} {
 #		cmd $sock $varName $errorString
 #		Where the errorString is only passed if an error occurred.
 #
+# Results:
+#	None
+#
 # Side Effects:
-#	This schedules a readable fileevent to read all the POST data
-#	asynchronously.  The data is appened to the named variable.
-#	The callback is made 
-
-proc Httpd_GetPostDataAsync {sock varName blockSize cmd} {
-    fileevent $sock readable \
-	[list HttpdReadPostGlobal $sock $varName $blockSize $cmd]
-    return
-}
-
-proc HttpdReadPostGlobal {sock varName blockSize {cmd {}}} {
-
-    # This fileevent callback can only access a global variable.
-    # But HttpdReadPost needs to affect a local variable in its
-    # caller so it can be shared with Httpd_GetPostData.
-    # So, the fileevent case has an extra procedure frame.
-
-    upvar #0 $varName buffer
-    HttpdReadPost $sock buffer $blockSize $cmd
-}
+#	Consumes post data and appends it to a variable.
 
 proc HttpdReadPost {sock varName blockSize {cmd {}}} {
     global Httpd
@@ -733,6 +817,7 @@ proc HttpdReadPost {sock varName blockSize {cmd {}}} {
 	    if {[info exist buffer]} {
 		append buffer $block
 	    }
+
 	    set data(count) [expr {$data(count) - [string length $block]}]
 	    if {$data(count) == 0} {
 		set doneMsg ""
@@ -741,10 +826,11 @@ proc HttpdReadPost {sock varName blockSize {cmd {}}} {
     }
     if {[info exist doneMsg]} {
 	Url_PostHook $sock 0
+	catch {fileevent $sock readable {}}
+	Httpd_Resume $sock
 	if {[string length $cmd]} {
 	    eval $cmd [list $sock $varName $doneMsg]
 	}
-	catch {fileevent $sock readable {}}
 	return $doneMsg
     } else {
 	return ""
@@ -844,6 +930,49 @@ proc HttpdCloseP {sock} {
     return $close
 }
 
+# Httpd_CompletionCallback --
+#
+#	Register a procedure to be called when an HTTP request is
+#	completed, either normally or forcibly closed.  This gives a
+#	URL implementation a guaranteed callback to clean up or log
+#	requests.
+#
+# Arguments:
+#	sock	The connection handle
+#	cmd	The callback to make.  These arguments are added:
+#		sock - the connection
+#		errmsg - An empty string, or an error message.
+#
+# Side Effects:
+# 	Registers the callback
+
+proc Httpd_CompletionCallback {sock cmd} {
+    upvar #0 Httpd$sock data
+    set data(callback) $cmd
+}
+
+# HttpdDoCallback --
+#
+#	Invoke the completion callback.
+#
+# Arguments:
+#	sock	The connection handle
+#	errmsg	The empty string, or an error message.
+#
+# Side Effects:
+# 	Invokes the callback
+
+proc HttpdDoCallback {sock {errmsg {}}} {
+    upvar #0 Httpd$sock data
+    if {[info exists data(callback)]} {
+	catch {eval $data(callback) {$sock $errmsg}}
+
+	# Ensure it is only called once
+	unset data(callback)
+    }
+    CountStop serviceTime $sock
+}
+
 # HttpdRespondHeader --
 #
 #	Utility routine for outputting response headers for normal data Does
@@ -861,11 +990,11 @@ proc HttpdCloseP {sock} {
 # 	Outputs header lines
 
 proc HttpdRespondHeader {sock type close size {code 200}} {
-    global Httpd Httpd_Errors
+    global Httpd
     upvar #0 Httpd$sock data
 
     set data(code) $code
-    append reply "HTTP/$data(version) $code $Httpd_Errors($code)" \n
+    append reply "HTTP/$data(version) $code [HttpdErrorString $code]" \n
     append reply "Date: [HttpdDate [clock seconds]]" \n
     append reply "Server: $Httpd(server)\n"
 
@@ -881,6 +1010,28 @@ proc HttpdRespondHeader {sock type close size {code 200}} {
     puts -nonewline $sock $reply
 }
 
+# HttpdErrorString --
+#
+#	Map from an error code to a meaningful string.
+#
+# Arguments:
+#	code	An HTTP error code, e.g., 200 or 404
+#
+# Results:
+#	An error string, e.g. "Data follows" or "File Not Found"
+#
+# Side Effects:
+# 	None
+
+proc HttpdErrorString { code } {
+    global Httpd_Errors
+    if {[info exist Httpd_Errors($code)]} {
+	return $Httpd_Errors($code)
+    } else {
+	return "Error $code"
+    }
+}
+
 # Httpd_SetCookie
 #	Define a cookie to be used in a reply
 #	Call this before using Httpd_ReturnFile or
@@ -942,23 +1093,27 @@ proc Httpd_ReturnFile {sock type path {offset 0}} {
 
     incr data(file_size) -$offset
 
-    set close [HttpdCloseP $sock]
-    HttpdRespondHeader $sock $type $close $data(file_size) 200
-    HttpdSetCookie $sock
-    puts $sock "Last-Modified: [HttpdDate [file mtime $path]]"
-    puts $sock ""
-    if {$data(proto) != "HEAD"} {
-	set in [open $path]		;# checking should already be done
-	fconfigure $in -translation binary -blocking 1
-	if {$offset != 0} {
-	    seek $in $offset
+    if {[catch {
+	set close [HttpdCloseP $sock]
+	HttpdRespondHeader $sock $type $close $data(file_size) 200
+	HttpdSetCookie $sock
+	puts $sock "Last-Modified: [HttpdDate [file mtime $path]]"
+	puts $sock ""
+	if {$data(proto) != "HEAD"} {
+	    set in [open $path]		;# checking should already be done
+	    fconfigure $in -translation binary -blocking 1
+	    if {$offset != 0} {
+		seek $in $offset
+	    }
+	    fconfigure $sock -translation binary -blocking $Httpd(sockblock)
+	    set data(infile) $in
+	    Httpd_Suspend $sock 0
+	    fcopy $in $sock -command [list HttpdCopyDone $in $sock $close]
+	} else {
+	    Httpd_SockClose $sock $close
 	}
-	fconfigure $sock -translation binary -blocking $Httpd(sockblock)
-	set data(infile) $in
-	Httpd_Suspend $sock 0
-	fcopy $in $sock -command [list HttpdCopyDone $in $sock $close]
-    } else {
-	Httpd_SockClose $sock $close
+    } err]} {
+	HttpdCloseFinal $sock $err
     }
 }
 
@@ -974,7 +1129,7 @@ proc Httpd_ReturnFile {sock type path {offset 0}} {
 #	Send the data down the socket
 
 proc Httpd_ReturnData {sock type content {code 200} {close 0}} {
-    global Httpd Httpd_Errors
+    global Httpd
     upvar #0 Httpd$sock data
 
     if {[Thread_Respond $sock \
@@ -986,14 +1141,18 @@ proc Httpd_ReturnData {sock type content {code 200} {close 0}} {
     if {$close == 0} {
     	set close [HttpdCloseP $sock]
     }
-    HttpdRespondHeader $sock $type $close [string length $content] $code
-    HttpdSetCookie $sock
-    puts $sock ""
-    if {$data(proto) != "HEAD"} {
-	fconfigure $sock -translation binary -blocking $Httpd(sockblock)
-	puts -nonewline $sock $content
+    if {[catch {
+	HttpdRespondHeader $sock $type $close [string length $content] $code
+	HttpdSetCookie $sock
+	puts $sock ""
+	if {$data(proto) != "HEAD"} {
+	    fconfigure $sock -translation binary -blocking $Httpd(sockblock)
+	    puts -nonewline $sock $content
+	}
+	Httpd_SockClose $sock $close
+    } err]} {
+	HttpdCloseFinal $sock $err
     }
-    Httpd_SockClose $sock $close
 }
 
 # Httpd_ReturnCacheableData
@@ -1011,7 +1170,7 @@ proc Httpd_ReturnData {sock type content {code 200} {close 0}} {
 #	Send the data down the socket
 
 proc Httpd_ReturnCacheableData {sock type content date {code 200}} {
-    global Httpd Httpd_Errors
+    global Httpd 
     upvar #0 Httpd$sock data
 
     if {[Thread_Respond $sock \
@@ -1021,26 +1180,57 @@ proc Httpd_ReturnCacheableData {sock type content date {code 200}} {
 
     Count urlreply
     set close [HttpdCloseP $sock]
-    HttpdRespondHeader $sock $type $close [string length $content] $code
-    HttpdSetCookie $sock
-    puts $sock "Last-Modified: [HttpdDate $date]"
-    puts $sock ""
-    if {$data(proto) != "HEAD"} {
-	fconfigure $sock -translation binary -blocking $Httpd(sockblock)
-	puts -nonewline $sock $content
+    if {[catch {
+	HttpdRespondHeader $sock $type $close [string length $content] $code
+	HttpdSetCookie $sock
+	puts $sock "Last-Modified: [HttpdDate $date]"
+	puts $sock ""
+	if {$data(proto) != "HEAD"} {
+	    fconfigure $sock -translation binary -blocking $Httpd(sockblock)
+	    puts -nonewline $sock $content
+	}
+	Httpd_SockClose $sock $close
+    } err]} {
+	HttpdCloseFinal $sock $err
     }
-    Httpd_SockClose $sock $close
 }
 
 # HttpdCopyDone -- this is used with fcopy when the copy completes.
 # Note that tcl8.0b1 had a bug in that errors during fcopy called
 # bgerror instead of this routine, which causes leaks.  Don't use b1.
+#
+# Arguments:
+#	in	Input channel, typically a file
+#	sock	Socket connection
+#	close	If true, the socket is closed after the copy.
+#	bytes	How many bytes were copied
+#	error	Optional error string.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	See Httpd_SockClose
 
 proc HttpdCopyDone {in sock close bytes {error {}}} {
-    Httpd_SockClose $sock $close
+    if {[string length $error] == 0} {
+	set error Close	;# Normal close, which triggers log message
+    }
+    Httpd_SockClose $sock $close $error
 }
 
+# HttpdCancel --
+#
 # Cancel a transaction if the client doesn't complete the request fast enough.
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Terminates the connection by returning an error page.
 
 proc HttpdCancel {sock} {
     upvar #0 Httpd$sock data
@@ -1056,9 +1246,22 @@ set Httpd_ErrorFormat {
     while trying to obtain <b>%3$s</b>.
 }
 
+# Httpd_Error --
+#
 # send the error message, log it, and close the socket.
 # Note that the Doc module tries to present a more palatable
 # error display page, but falls back to this if necessary.
+#
+# Arguments:
+#	sock	Socket connection
+#	code	HTTP error code, e.g., 500
+#	detail  Optional string to append to standard error message.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Generates a HTTP response.
 
 proc Httpd_Error {sock code {detail ""}} {
 
@@ -1068,18 +1271,11 @@ proc Httpd_Error {sock code {detail ""}} {
     }
 
     upvar #0 Httpd$sock data
-    global Httpd Httpd_Errors Httpd_ErrorFormat
+    global Httpd Httpd_ErrorFormat
 
     Count errors
     append data(url) ""
-    if {[info exists data(code)]} {
-	set detail "OLDCODE $code $detail"
-    }
-    if {[info exists data(infile)]} {
-	set detail "BUSY $detail"
-    }
-    set data(code) $code
-    set message [format $Httpd_ErrorFormat $code $Httpd_Errors($code)  $data(url)]
+    set message [format $Httpd_ErrorFormat $code [HttpdErrorString $code] $data(url)]
     append message <br>$detail
     if {$code == 500} {
 	append message "<h2>Tcl Call Trace</h2>"
@@ -1088,6 +1284,13 @@ proc Httpd_Error {sock code {detail ""}} {
 	}
     }
     Log $sock Error $code $data(url) $detail
+
+    # We know something is bad here, so we make the completion callback
+    # and then unregister it so we don't get an extra call as a side
+    # effect of trying to reply.
+
+    HttpdDoCallback $sock $message
+
     if {[info exists data(infile)]} {
 	# We've already started a reply, so just bail out
 	Httpd_SockClose $sock 1
@@ -1112,11 +1315,23 @@ set HttpdRedirectFormat {
     </body></html>
 }
 
+# Httpd_Redirect --
+#
 # Generate a redirect reply (code 302)
+#
+# Arguments:
+#	newurl	New URL to redirect to.
+#	sock	Socket connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Generates an HTTP reply
 
 proc Httpd_Redirect {newurl sock} {
     upvar #0 Httpd$sock data
-    global Httpd Httpd_Errors HttpdRedirectFormat
+    global Httpd HttpdRedirectFormat
 
     if {[Thread_Respond $sock \
 	    [list Httpd_Redirect $newurl $sock]]} {
@@ -1138,18 +1353,43 @@ proc Httpd_Redirect {newurl sock} {
     Httpd_SockClose $sock $close
 }
 
+# Httpd_RedirectSelf --
+#
 # Generate a redirect to another URL on this server.
+#
+# Arguments:
+#	newurl	Server-relative URL to redirect to.
+#	sock	Socket connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Generates an HTTP reply.
 
 proc Httpd_RedirectSelf {newurl sock} {
     Httpd_Redirect [Httpd_SelfUrl $newurl $sock] $sock
 }
 
+# Httpd_SelfUrl --
+#
 # Create an absolute URL for this server
+#
+# Arguments:
+#	url	A server-relative URL on this server.
+#	sock	The current connection so we can tell if it
+#		is the regular port or the secure port.
+#
+# Results:
+#	An absolute URL.
+#
+# Side Effects:
+#	None
 
 proc Httpd_SelfUrl {url {sock ""}} {
-    global Httpd env
+    global Httpd
     if {$sock == ""} {
-	set sock $env(HTTP_CHANNEL)
+	set sock $Httpd(currentSocket)
     }
     upvar #0 Httpd$sock data
 
@@ -1180,39 +1420,123 @@ proc Httpd_SelfUrl {url {sock ""}} {
     append newurl $url
 }
 
+# Httpd_Protocol --
+#
 # Return the protocol for the connection
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	Either "http" or "https"
+#
+# Side Effects:
+#	None
 
 proc Httpd_Protocol {sock} {
     upvar #0 Httpd$sock data
     return [lindex $data(self) 0]
 }
 
+# Httpd_Name --
+#
 # Return the server name for the connection
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	The name of the server, e.g., www.tcltk.org
+#
+# Side Effects:
+#	None
 
 proc Httpd_Name {sock} {
     upvar #0 Httpd$sock data
     return [lindex $data(self) 1]
 }
 
+# Httpd_Port --
+#
 # Return the port for the connection
+#
+# Arguments:
+#	sock	The current connection. If empty, then the
+#		regular (non-secure) port is returned.  Otherwise
+#		the port of this connection is returned.
+#
+# Results:
+#	The server's port.
+#
+# Side Effects:
+#	None
 
-proc Httpd_Port {sock} {
-    upvar #0 Httpd$sock data
-    return [lindex $data(self) 2]
+proc Httpd_Port {{sock {}}} {
+    if {[string length $sock]} {
+
+	# Return the port for this connection
+
+	upvar #0 Httpd$sock data
+	return [lindex $data(self) 2]
+    } else {
+
+	# Return the non-secure listening port
+
+	global Httpd
+	if {[info exist Httpd(port)]} {
+	    return $Httpd(port)
+	} else {
+	    return {}
+	}
+    }
+}
+# Httpd_SecurePort --
+#
+#	Return the secure port of this server
+#
+# Arguments:
+#
+# Results:
+#	The server's secure port.
+#
+# Side Effects:
+#	None
+
+proc Httpd_SecurePort {} {
+
+    # Return the secure listening port
+
+    global Httpd
+    if {[info exist Httpd(https_port)]} {
+	return $Httpd(https_port)
+    } else {
+	return {}
+    }
 }
 
+
+# Httpd_RedirectDir --
+#
 # Generate a redirect because the trailing slash isn't present
 # on a URL that corresponds to a directory.
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Generate a redirect page.
 
 proc Httpd_RedirectDir {sock} {
     global Httpd
     upvar #0 Httpd$sock data
-    # set url http://$Httpd(name)
-    # if {$Httpd(port) != 80} {
-    # 	append url :$Httpd(port)
-    #    }
-    # Httpd_Redirect $url$data(url)/ $sock
-    Httpd_Redirect $data(url)/ $sock
+    set url  $data(url)/
+    if {[info exist data(query)] && [string length $data(query)]} {
+	append url ?$data(query)
+    }
+    Httpd_Redirect $url $sock
 }
 
 set HttpdAuthorizationFormat {
@@ -1229,13 +1553,24 @@ set HttpdAuthorizationFormat {
     </BODY></HTML>
 }
 
+# Httpd_RequestAuth --
+#
 # Generate the (401) Authorization required reply
-# type - usually "Basic"
-# realm - browsers use this to cache credentials
+#
+# Arguments:
+#	sock	Socket connection
+#	type	usually "Basic"
+#	realm	browsers use this to cache credentials
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Generate an authorization challenge response.
 
 proc Httpd_RequestAuth {sock type realm} {
     upvar #0 Httpd$sock data
-    global Httpd Httpd_Errors HttpdAuthorizationFormat
+    global Httpd HttpdAuthorizationFormat
 
     if {[Thread_Respond $sock \
 	    [list Httpd_RequestAuth $sock $type $realm]]} {
@@ -1250,7 +1585,18 @@ proc Httpd_RequestAuth {sock type realm} {
     Httpd_SockClose $sock $close
 }
 
+# HttpdDate --
+#
 # generate a date string in HTTP format
+#
+# Arguments:
+#	seconds	Clock seconds value
+#
+# Results:
+#	A date string.
+#
+# Side Effects:
+#	None
 
 proc HttpdDate {seconds} {
     return [clock format $seconds -format {%a, %d %b %Y %T GMT} -gmt true]
@@ -1278,7 +1624,22 @@ proc Httpd_SockClose {sock closeit {message Close}} {
 
     if {[string length $message]} {
 	Log $sock $message
+	if {$message == "Close"} {
+
+	    # This is a normal close.  Any other message
+	    # is some sort of error.
+
+	    set message ""
+	}
     }
+    # Call back to the URL domain implementation so they are
+    # sure to see the end of all their HTTP transactions.
+    # There is a slight chance of an error reading any un-read
+    # Post data, but if the URL domain didn't want to read it,
+    # then they obviously don't care.
+
+    HttpdDoCallback $sock $message
+
     Count connections -1
     if {[info exist data(infile)]} {
 
@@ -1297,9 +1658,9 @@ proc Httpd_SockClose {sock closeit {message Close}} {
 		after cancel $data(cancel)
 	    }
 	    set data(cancel) [after $Httpd(timeout3) \
-		[list HttpdCloseFinal $sock]]
+		[list HttpdCloseFinal $sock "timeout reading extra POST data"]]
 
-	    Httpd_GetPostDataAsync $sock "" $data(count) HttpdCloseFinal
+	    Httpd_GetPostDataAsync $sock "" $data(count) HttpdReadPostDone
 	} else {
 	    HttpdCloseFinal $sock
 	}
@@ -1308,7 +1669,44 @@ proc Httpd_SockClose {sock closeit {message Close}} {
     }
 }
 
-proc HttpdCloseFinal {sock args} {
+# HttpdReadPostDone --
+#
+#	Callback is made to this when we are done cleaning up any
+#	unread post data.
+#
+# Arguments:
+#	sock	Socket connection
+#	varname	Name of variable with post data
+#	errmsg	If not empty, an error occurred on the socket.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	None here, but see HttpdCloseFinal.
+
+proc HttpdReadPostDone {sock var errmsg} {
+    HttpdCloseFinal $sock $errmsg
+}
+
+# HttpdCloseFinal --
+#
+#	Central close procedure.  All close operations should funnel
+#	through here so that the right cleanup occurs.
+#
+# Arguments:
+#	sock	Socket connection
+#	errmsg	If non-empty, then something went wrong on the socket.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Cleans up any after event associated with the connection.
+#	Closes the socket.
+#	Makes the callback to the URL domain implementation.
+
+proc HttpdCloseFinal {sock {errmsg {}}} {
     upvar #0 Httpd$sock data
     Count sockets -1
     if {[info exists data(cancel)]} {
@@ -1316,8 +1714,41 @@ proc HttpdCloseFinal {sock args} {
     }
     if {[catch {close $sock} err]} {
 	Log $sock CloseError $err
+	if {[string length $errmsg] == 0} {
+	    set errmsg $err
+	}
+    }
+    HttpdDoCallback $sock $errmsg
+    if {[info exist data]} {
+	unset data
+    }
+}
+
+# Httpd_RequestComplete --
+#
+#	Detect if a request has been sent.  The holder of a socket
+#	might need to know of the URL request was completed with
+#	one of the return-data commands, or is still lingering open.
+#
+# Arguments:
+#	sock	Socket connection
+#
+# Results:
+#	1 if the request was completed, 0 otherwise.
+#
+# Side Effects:
+#	None.
+
+proc Httpd_RequestComplete {sock} {
+    upvar #0 Httpd$sock data
+    if {![info exist data(state)] || $data(state) == "start"} {
+	
+	# The connection was closed or reset in a keep-alive situation.
+
+	return 1
+    } else {
+	return 0
     }
-    unset data
 }
 
 # server specific version of bgerror - indent to hide from indexer
@@ -1331,6 +1762,20 @@ proc HttpdCloseFinal {sock args} {
 	}
     }
 
+# HttpdCookieLog --
+#
+#	Special loggin procedure to debug cookie implementation.
+#
+# Arguments:
+#	sock	Socket connection
+#	what	What procedure is doing the logging.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Writes to the debug log.
+
 proc HttpdCookieLog {sock what} {
     global Log Httpd
     upvar #0 Httpd$sock data
@@ -1358,8 +1803,21 @@ proc HttpdCookieLog {sock what} {
     }
 }
 
-# Suspend Wire Callback - for async transactions
-# Use HttpdReset once you are back in business
+# Httpd_Suspend --
+#
+#	Suspend Wire Callback - for async transactions
+#	Use Httpd_Resume once you are back in business
+#
+# Arguments:
+#	sock	Socket connection
+#	timeout Timeout period.  After this the request is aborted.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Disables fileevents and sets up a timer.
+
 proc Httpd_Suspend {sock {timeout ""}} {
     global Httpd
     upvar #0 Httpd$sock data
@@ -1371,6 +1829,9 @@ proc Httpd_Suspend {sock {timeout ""}} {
 	after cancel $data(cancel)
 	unset data(cancel)
     }
+    if {[info exists Httpd(currentSocket)]} {
+	unset Httpd(currentSocket)
+    }
     if {$timeout == ""} {
 	set timeout $Httpd(timeout2)
     }
@@ -1379,10 +1840,73 @@ proc Httpd_Suspend {sock {timeout ""}} {
     }
 }
 
+# Httpd_Resume --
+#
+#	Resume processing of a request.  Sets up a bit of global state that
+#	has been cleared by Httpd_Suspend.
+#
+# Arguments:
+#	sock	Socket connection
+#	timeout Timeout period.  After this the request is aborted.
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Restores the Httpd(currentSocket) setting.
+
+proc Httpd_Resume {sock {timeout ""}} {
+    upvar #0 Httpd$sock data
+    global Httpd
+    set Httpd(currentSocket) $sock
+    if {[info exists data(cancel)]} {
+	after cancel $data(cancel)
+    }
+    if {$timeout == ""} {
+	set timeout $Httpd(timeout1)
+    }
+    set data(cancel) [after $timeout \
+	[list Httpd_SockClose $sock 1 "timeout"]]
+}
+
+# Httpd_CurrentSocket --
+#
+#	Return (or set) the handle to the current socket.
+#
+# Arguments:
+#	sock	Optional - if specified, set the current socket.
+#
+# Results:
+#	A socket.
+#
+# Side Effects:
+#	None.
+
+proc Httpd_CurrentSocket {{sock {}}} {
+    global Httpd
+    if {[string length $sock]} {
+	set Httpd(currentSocket) $sock
+    }
+    return $Httpd(currentSocket)
+}
+
+# Httpd_Pair --
+#
 #
 # Pair two fd's - typically for tunnelling
 # Close both if either one closes (or gets an error)
 #
+#
+# Arguments:
+#	sock	Socket connection
+#	fd	Any other I/O connection
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Sets up fileevents for proxy'ing data.
+
 proc Httpd_Pair {sock fd} {
     global Httpd
     upvar #0 Httpd$sock data
@@ -1398,6 +1922,21 @@ proc Httpd_Pair {sock fd} {
     fileevent $fd readable [list HttpdReflect $fd $sock]
 }
 
+# HttpdReflect --
+#
+#	This is logically fcopy in both directions, but the core
+#	prevents us from doing that so we do it by hand.
+#
+# Arguments:
+#	in	Input channel
+#	out	Output channel
+#
+# Results:
+#	None
+#
+# Side Effects:
+#	Copy data between channels.
+
 proc HttpdReflect {in out} {
     global Httpd
     if {[catch {
